package com.sun.org.apache.xerces.internal.dom;

import com.sun.org.apache.xerces.internal.util.AugmentationsImpl;
import java.util.Vector;
import com.sun.org.apache.xerces.internal.xni.parser.XMLDocumentSource;
import com.sun.org.apache.xerces.internal.xs.ElementPSVI;
import com.sun.org.apache.xerces.internal.xni.XMLResourceIdentifier;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import com.sun.org.apache.xerces.internal.impl.dv.XSSimpleType;
import com.sun.org.apache.xerces.internal.xs.AttributePSVI;
import com.sun.org.apache.xerces.internal.xni.XMLAttributes;
import com.sun.org.apache.xerces.internal.xni.XNIException;
import com.sun.org.apache.xerces.internal.xni.Augmentations;
import com.sun.org.apache.xerces.internal.xni.XMLLocator;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Entity;
import org.w3c.dom.NamedNodeMap;
import com.sun.org.apache.xerces.internal.impl.dtd.XMLDTDLoader;
import org.w3c.dom.Element;
import org.w3c.dom.DocumentType;
import java.io.IOException;
import com.sun.org.apache.xerces.internal.util.XML11Char;
import com.sun.org.apache.xerces.internal.util.XMLChar;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;
import org.w3c.dom.Comment;
import org.w3c.dom.Attr;
import com.sun.org.apache.xerces.internal.impl.xs.util.SimpleLocator;
import com.sun.org.apache.xerces.internal.xni.parser.XMLComponent;
import com.sun.org.apache.xerces.internal.impl.Constants;
import com.sun.org.apache.xerces.internal.util.XMLSymbols;
import java.util.ArrayList;
import com.sun.org.apache.xerces.internal.util.NamespaceSupport;
import com.sun.org.apache.xerces.internal.impl.dtd.XMLDTDValidator;
import com.sun.org.apache.xerces.internal.xni.XMLString;
import org.w3c.dom.Node;
import java.util.List;
import com.sun.org.apache.xerces.internal.xni.NamespaceContext;
import org.w3c.dom.DOMErrorHandler;
import com.sun.org.apache.xerces.internal.util.SymbolTable;
import com.sun.org.apache.xerces.internal.impl.RevalidationHandler;
import com.sun.org.apache.xerces.internal.xni.QName;
import com.sun.org.apache.xerces.internal.xni.XMLDocumentHandler;

public class DOMNormalizer implements XMLDocumentHandler
{
    protected static final boolean DEBUG_ND = false;
    protected static final boolean DEBUG = false;
    protected static final boolean DEBUG_EVENTS = false;
    protected static final String PREFIX = "NS";
    protected DOMConfigurationImpl fConfiguration;
    protected CoreDocumentImpl fDocument;
    protected final DOMNormalizer.XMLAttributesProxy fAttrProxy;
    protected final QName fQName;
    protected RevalidationHandler fValidationHandler;
    protected SymbolTable fSymbolTable;
    protected DOMErrorHandler fErrorHandler;
    private final DOMErrorImpl fError;
    protected boolean fNamespaceValidation;
    protected boolean fPSVI;
    protected final NamespaceContext fNamespaceContext;
    protected final NamespaceContext fLocalNSBinder;
    protected final List<Node> fAttributeList;
    protected final DOMLocatorImpl fLocator;
    protected Node fCurrentNode;
    private final QName fAttrQName;
    final XMLString fNormalizedValue;
    private XMLDTDValidator fDTDValidator;
    public static final XMLString EMPTY_STRING;
    private boolean fAllWhitespace;
    
    public DOMNormalizer() {
        this.fConfiguration = null;
        this.fDocument = null;
        this.fAttrProxy = new DOMNormalizer.XMLAttributesProxy();
        this.fQName = new QName();
        this.fError = new DOMErrorImpl();
        this.fNamespaceValidation = false;
        this.fPSVI = false;
        this.fNamespaceContext = new NamespaceSupport();
        this.fLocalNSBinder = new NamespaceSupport();
        this.fAttributeList = new ArrayList(5);
        this.fLocator = new DOMLocatorImpl();
        this.fCurrentNode = null;
        this.fAttrQName = new QName();
        this.fNormalizedValue = new XMLString(new char[16], 0, 0);
        this.fAllWhitespace = false;
    }
    
    protected void normalizeDocument(CoreDocumentImpl document, DOMConfigurationImpl config) {
        this.fDocument = document;
        this.fConfiguration = config;
        this.fAllWhitespace = false;
        this.fNamespaceValidation = false;
        String xmlVersion = this.fDocument.getXmlVersion();
        String schemaType = null;
        String[] schemaLocations = null;
        this.fSymbolTable = (SymbolTable)this.fConfiguration.getProperty("http://apache.org/xml/properties/internal/symbol-table");
        this.fNamespaceContext.reset();
        this.fNamespaceContext.declarePrefix(XMLSymbols.EMPTY_STRING, null);
        if ((this.fConfiguration.features & 0x40) != 0x0) {
            String schemaLang = (String)this.fConfiguration.getProperty("http://java.sun.com/xml/jaxp/properties/schemaLanguage");
            if (schemaLang != null && schemaLang.equals(Constants.NS_XMLSCHEMA)) {
                schemaType = "http://www.w3.org/2001/XMLSchema";
                this.fValidationHandler = CoreDOMImplementationImpl.singleton.getValidator(schemaType, xmlVersion);
                this.fConfiguration.setFeature("http://apache.org/xml/features/validation/schema", true);
                this.fConfiguration.setFeature("http://apache.org/xml/features/validation/schema-full-checking", true);
                this.fNamespaceValidation = true;
                this.fPSVI = ((this.fConfiguration.features & 0x80) != 0x0);
            }
            else {
                schemaType = "http://www.w3.org/TR/REC-xml";
                if (schemaLang != null) {
                    schemaLocations = (String[])(String[])this.fConfiguration.getProperty("http://java.sun.com/xml/jaxp/properties/schemaSource");
                }
                this.fConfiguration.setDTDValidatorFactory(xmlVersion);
                this.fValidationHandler = CoreDOMImplementationImpl.singleton.getValidator(schemaType, xmlVersion);
                this.fPSVI = false;
            }
            this.fConfiguration.setFeature("http://xml.org/sax/features/validation", true);
            this.fDocument.clearIdentifiers();
            if (this.fValidationHandler != null) {
                ((XMLComponent)this.fValidationHandler).reset(this.fConfiguration);
            }
        }
        else {
            this.fValidationHandler = null;
        }
        this.fErrorHandler = (DOMErrorHandler)this.fConfiguration.getParameter("error-handler");
        if (this.fValidationHandler != null) {
            this.fValidationHandler.setDocumentHandler(this);
            this.fValidationHandler.startDocument(new SimpleLocator(this.fDocument.fDocumentURI, this.fDocument.fDocumentURI, -1, -1), this.fDocument.encoding, this.fNamespaceContext, null);
            this.fValidationHandler.xmlDecl(this.fDocument.getXmlVersion(), this.fDocument.getXmlEncoding(), this.fDocument.getXmlStandalone() ? "yes" : "no", null);
        }
        try {
            if (schemaType == "http://www.w3.org/TR/REC-xml") {
                this.processDTD(xmlVersion, (schemaLocations != null) ? schemaLocations[0] : null);
            }
            Node next;
            for (Node kid = this.fDocument.getFirstChild(); kid != null; kid = next) {
                next = kid.getNextSibling();
                kid = this.normalizeNode(kid);
                if (kid != null) {
                    next = kid;
                }
            }
            if (this.fValidationHandler != null) {
                this.fValidationHandler.endDocument(null);
                this.fValidationHandler.setDocumentHandler(null);
                CoreDOMImplementationImpl.singleton.releaseValidator(schemaType, xmlVersion, this.fValidationHandler);
                this.fValidationHandler = null;
            }
        }
        catch (RuntimeException e) {
            if (this.fValidationHandler != null) {
                this.fValidationHandler.setDocumentHandler(null);
                CoreDOMImplementationImpl.singleton.releaseValidator(schemaType, xmlVersion, this.fValidationHandler);
                this.fValidationHandler = null;
            }
            if ((e instanceof AbortException)) {
                return;
            }
            throw e;
        }
    }
    
    protected Node normalizeNode(Node node) {
        int type = node.getNodeType();
        this.fLocator.fRelatedNode = node;
        switch (type) {
            case 10: {
                break;
            }
            case 1: {
                if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0 && this.fDocument.isXMLVersionChanged()) {
                    boolean wellformed;
                    if (this.fNamespaceValidation) {
                        wellformed = CoreDocumentImpl.isValidQName(node.getPrefix(), node.getLocalName(), this.fDocument.isXML11Version());
                    }
                    else {
                        wellformed = CoreDocumentImpl.isXMLName(node.getNodeName(), this.fDocument.isXML11Version());
                    }
                    if (!wellformed) {
                        String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "wf-invalid-character-in-node-name", new Object[] { "Element", node.getNodeName() });
                        reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg, (short)2, "wf-invalid-character-in-node-name");
                    }
                }
                this.fNamespaceContext.pushContext();
                this.fLocalNSBinder.reset();
                ElementImpl elem = (ElementImpl)node;
                if (elem.needsSyncChildren()) {
                    elem.synchronizeChildren();
                }
                AttributeMap attributes = elem.hasAttributes() ? ((AttributeMap)elem.getAttributes()) : null;
                if ((this.fConfiguration.features & 0x1) != 0x0) {
                    this.namespaceFixUp(elem, attributes);
                    if ((this.fConfiguration.features & 0x200) == 0x0) {
                        if (attributes == null) {
                            attributes = (elem.hasAttributes() ? ((AttributeMap)elem.getAttributes()) : null);
                        }
                        if (attributes != null) {
                            for (int i = 0; i < attributes.getLength(); ++i) {
                                Attr att = (Attr)attributes.getItem(i);
                                if (XMLSymbols.PREFIX_XMLNS.equals(att.getPrefix()) || XMLSymbols.PREFIX_XMLNS.equals(att.getName())) {
                                    elem.removeAttributeNode(att);
                                    --i;
                                }
                            }
                        }
                    }
                }
                else if (attributes != null) {
                    for (int i = 0; i < attributes.getLength(); ++i) {
                        Attr attr = (Attr)attributes.item(i);
                        attr.normalize();
                        if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0) {
                            isAttrValueWF(this.fErrorHandler, this.fError, this.fLocator, attributes, attr, attr.getValue(), this.fDocument.isXML11Version());
                            if (this.fDocument.isXMLVersionChanged()) {
                                boolean wellformed;
                                if (this.fNamespaceValidation) {
                                    wellformed = CoreDocumentImpl.isValidQName(node.getPrefix(), node.getLocalName(), this.fDocument.isXML11Version());
                                }
                                else {
                                    wellformed = CoreDocumentImpl.isXMLName(node.getNodeName(), this.fDocument.isXML11Version());
                                }
                                if (!wellformed) {
                                    String msg2 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "wf-invalid-character-in-node-name", new Object[] { "Attr", node.getNodeName() });
                                    reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg2, (short)2, "wf-invalid-character-in-node-name");
                                }
                            }
                        }
                    }
                }
                if (this.fValidationHandler != null) {
                    this.fAttrProxy.setAttributes(attributes, this.fDocument, elem);
                    this.updateQName(elem, this.fQName);
                    this.fConfiguration.fErrorHandlerWrapper.fCurrentNode = node;
                    this.fCurrentNode = node;
                    this.fValidationHandler.startElement(this.fQName, this.fAttrProxy, null);
                }
                Node next;
                for (Node kid = elem.getFirstChild(); kid != null; kid = next) {
                    next = kid.getNextSibling();
                    kid = this.normalizeNode(kid);
                    if (kid != null) {
                        next = kid;
                    }
                }
                if (this.fValidationHandler != null) {
                    this.updateQName(elem, this.fQName);
                    this.fConfiguration.fErrorHandlerWrapper.fCurrentNode = node;
                    this.fCurrentNode = node;
                    this.fValidationHandler.endElement(this.fQName, null);
                }
                this.fNamespaceContext.popContext();
                break;
            }
            case 8: {
                if ((this.fConfiguration.features & 0x20) == 0x0) {
                    Node prevSibling = node.getPreviousSibling();
                    Node parent = node.getParentNode();
                    parent.removeChild(node);
                    if (prevSibling != null && prevSibling.getNodeType() == 3) {
                        Node nextSibling = prevSibling.getNextSibling();
                        if (nextSibling != null && nextSibling.getNodeType() == 3) {
                            ((TextImpl)nextSibling).insertData(0, prevSibling.getNodeValue());
                            parent.removeChild(prevSibling);
                            return nextSibling;
                        }
                    }
                    break;
                }
                else {
                    if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0) {
                        String commentdata = ((Comment)node).getData();
                        isCommentWF(this.fErrorHandler, this.fError, this.fLocator, commentdata, this.fDocument.isXML11Version());
                    }
                    if (this.fValidationHandler != null) {
                        this.fValidationHandler.comment(DOMNormalizer.EMPTY_STRING, null);
                        break;
                    }
                    break;
                }
            }
            case 5: {
                if ((this.fConfiguration.features & 0x4) == 0x0) {
                    Node prevSibling = node.getPreviousSibling();
                    Node parent = node.getParentNode();
                    ((EntityReferenceImpl)node).setReadOnly(false, true);
                    this.expandEntityRef(parent, node);
                    parent.removeChild(node);
                    Node next2 = (prevSibling != null) ? prevSibling.getNextSibling() : parent.getFirstChild();
                    if (prevSibling != null && next2 != null && prevSibling.getNodeType() == 3 && next2.getNodeType() == 3) {
                        return prevSibling;
                    }
                    return next2;
                }
                else {
                    if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0 && this.fDocument.isXMLVersionChanged()) {
                        CoreDocumentImpl.isXMLName(node.getNodeName(), this.fDocument.isXML11Version());
                        break;
                    }
                    break;
                }
            }
            case 4: {
                if ((this.fConfiguration.features & 0x8) == 0x0) {
                    Node prevSibling = node.getPreviousSibling();
                    if (prevSibling != null && prevSibling.getNodeType() == 3) {
                        ((Text)prevSibling).appendData(node.getNodeValue());
                        node.getParentNode().removeChild(node);
                        return prevSibling;
                    }
                    Text text = this.fDocument.createTextNode(node.getNodeValue());
                    Node parent2 = node.getParentNode();
                    node = parent2.replaceChild(text, node);
                    return text;
                }
                else {
                    if (this.fValidationHandler != null) {
                        this.fConfiguration.fErrorHandlerWrapper.fCurrentNode = node;
                        this.fCurrentNode = node;
                        this.fValidationHandler.startCDATA(null);
                        this.fValidationHandler.characterData(node.getNodeValue(), null);
                        this.fValidationHandler.endCDATA(null);
                    }
                    String value = node.getNodeValue();
                    if ((this.fConfiguration.features & 0x10) != 0x0) {
                        Node parent2 = node.getParentNode();
                        if (this.fDocument.errorChecking) {
                            isXMLCharWF(this.fErrorHandler, this.fError, this.fLocator, node.getNodeValue(), this.fDocument.isXML11Version());
                        }
                        int index;
                        while ((index = value.indexOf("]]>")) >= 0) {
                            node.setNodeValue(value.substring(0, index + 2));
                            value = value.substring(index + 2);
                            Node firstSplitNode = node;
                            Node newChild = this.fDocument.createCDATASection(value);
                            parent2.insertBefore(newChild, node.getNextSibling());
                            node = newChild;
                            this.fLocator.fRelatedNode = firstSplitNode;
                            String msg3 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "cdata-sections-splitted", (Object[])null);
                            reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg3, (short)1, "cdata-sections-splitted");
                        }
                        break;
                    }
                    else {
                        if (this.fDocument.errorChecking) {
                            isCDataWF(this.fErrorHandler, this.fError, this.fLocator, value, this.fDocument.isXML11Version());
                            break;
                        }
                        break;
                    }
                }
            }
            case 3: {
                Node next3 = node.getNextSibling();
                if (next3 != null && next3.getNodeType() == 3) {
                    ((Text)node).appendData(next3.getNodeValue());
                    node.getParentNode().removeChild(next3);
                    return node;
                }
                if (node.getNodeValue().length() == 0) {
                    node.getParentNode().removeChild(node);
                    break;
                }
                short nextType = (next3 != null) ? next3.getNodeType() : -1;
                if (nextType == -1 || (((this.fConfiguration.features & 0x4) != 0x0 || nextType != 6) && ((this.fConfiguration.features & 0x20) != 0x0 || nextType != 8) && ((this.fConfiguration.features & 0x8) != 0x0 || nextType != 4))) {
                    if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0) {
                        isXMLCharWF(this.fErrorHandler, this.fError, this.fLocator, node.getNodeValue(), this.fDocument.isXML11Version());
                    }
                    if (this.fValidationHandler != null) {
                        this.fConfiguration.fErrorHandlerWrapper.fCurrentNode = node;
                        this.fCurrentNode = node;
                        this.fValidationHandler.characterData(node.getNodeValue(), null);
                        if (!this.fNamespaceValidation) {
                            if (this.fAllWhitespace) {
                                this.fAllWhitespace = false;
                                ((TextImpl)node).setIgnorableWhitespace(true);
                            }
                            else {
                                ((TextImpl)node).setIgnorableWhitespace(false);
                            }
                        }
                    }
                }
                break;
            }
            case 7: {
                if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0) {
                    ProcessingInstruction pinode = (ProcessingInstruction)node;
                    String target = pinode.getTarget();
                    boolean wellformed;
                    if (this.fDocument.isXML11Version()) {
                        wellformed = XML11Char.isXML11ValidName(target);
                    }
                    else {
                        wellformed = XMLChar.isValidName(target);
                    }
                    if (!wellformed) {
                        String msg4 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "wf-invalid-character-in-node-name", new Object[] { "Element", node.getNodeName() });
                        reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg4, (short)2, "wf-invalid-character-in-node-name");
                    }
                    isXMLCharWF(this.fErrorHandler, this.fError, this.fLocator, pinode.getData(), this.fDocument.isXML11Version());
                }
                if (this.fValidationHandler != null) {
                    this.fValidationHandler.processingInstruction(((ProcessingInstruction)node).getTarget(), DOMNormalizer.EMPTY_STRING, null);
                    break;
                }
                break;
            }
        }
        return null;
    }
    
    private void processDTD(String xmlVersion, String schemaLocation) {
        String rootName = null;
        String publicId = null;
        String systemId = schemaLocation;
        String baseSystemId = this.fDocument.getDocumentURI();
        String internalSubset = null;
        DocumentType docType = this.fDocument.getDoctype();
        if (docType != null) {
            rootName = docType.getName();
            publicId = docType.getPublicId();
            if (systemId == null || systemId.length() == 0) {
                systemId = docType.getSystemId();
            }
            internalSubset = docType.getInternalSubset();
        }
        else {
            Element elem = this.fDocument.getDocumentElement();
            if (elem == null) {
                return;
            }
            rootName = elem.getNodeName();
            if (systemId == null || systemId.length() == 0) {
                return;
            }
        }
        XMLDTDLoader loader = null;
        try {
            this.fValidationHandler.doctypeDecl(rootName, publicId, systemId, null);
            loader = CoreDOMImplementationImpl.singleton.getDTDLoader(xmlVersion);
            loader.setFeature("http://xml.org/sax/features/validation", true);
            loader.setEntityResolver(this.fConfiguration.getEntityResolver());
            loader.setErrorHandler(this.fConfiguration.getErrorHandler());
            loader.loadGrammarWithContext((XMLDTDValidator)this.fValidationHandler, rootName, publicId, systemId, baseSystemId, internalSubset);
        }
        catch (IOException ex) {}
        finally {
            if (loader != null) {
                CoreDOMImplementationImpl.singleton.releaseDTDLoader(xmlVersion, loader);
            }
        }
    }
    
    protected final void expandEntityRef(Node parent, Node reference) {
        Node next;
        for (Node kid = reference.getFirstChild(); kid != null; kid = next) {
            next = kid.getNextSibling();
            parent.insertBefore(kid, reference);
        }
    }
    
    protected final void namespaceFixUp(ElementImpl element, AttributeMap attributes) {
        if (attributes != null) {
            for (int k = 0; k < attributes.getLength(); ++k) {
                Attr attr = (Attr)attributes.getItem(k);
                String uri = attr.getNamespaceURI();
                if (uri != null && uri.equals(NamespaceContext.XMLNS_URI)) {
                    String value = attr.getNodeValue();
                    if (value == null) {
                        value = XMLSymbols.EMPTY_STRING;
                    }
                    if (this.fDocument.errorChecking && value.equals(NamespaceContext.XMLNS_URI)) {
                        this.fLocator.fRelatedNode = attr;
                        String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "CantBindXMLNS", (Object[])null);
                        reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg, (short)2, "CantBindXMLNS");
                    }
                    else {
                        String prefix = attr.getPrefix();
                        prefix = ((prefix == null || prefix.length() == 0) ? XMLSymbols.EMPTY_STRING : this.fSymbolTable.addSymbol(prefix));
                        String localpart = this.fSymbolTable.addSymbol(attr.getLocalName());
                        if (prefix == XMLSymbols.PREFIX_XMLNS) {
                            value = this.fSymbolTable.addSymbol(value);
                            if (value.length() != 0) {
                                this.fNamespaceContext.declarePrefix(localpart, value);
                            }
                        }
                        else {
                            value = this.fSymbolTable.addSymbol(value);
                            this.fNamespaceContext.declarePrefix(XMLSymbols.EMPTY_STRING, (value.length() != 0) ? value : null);
                        }
                    }
                }
            }
        }
        String uri = element.getNamespaceURI();
        String prefix = element.getPrefix();
        if (uri != null) {
            uri = this.fSymbolTable.addSymbol(uri);
            prefix = ((prefix == null || prefix.length() == 0) ? XMLSymbols.EMPTY_STRING : this.fSymbolTable.addSymbol(prefix));
            if (this.fNamespaceContext.getURI(prefix) != uri) {
                this.addNamespaceDecl(prefix, uri, element);
                this.fLocalNSBinder.declarePrefix(prefix, uri);
                this.fNamespaceContext.declarePrefix(prefix, uri);
            }
        }
        else if (element.getLocalName() == null) {
            if (this.fNamespaceValidation) {
                String msg2 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "NullLocalElementName", new Object[] { element.getNodeName() });
                reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg2, (short)3, "NullLocalElementName");
            }
            else {
                String msg2 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "NullLocalElementName", new Object[] { element.getNodeName() });
                reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg2, (short)2, "NullLocalElementName");
            }
        }
        else {
            uri = this.fNamespaceContext.getURI(XMLSymbols.EMPTY_STRING);
            if (uri != null && uri.length() > 0) {
                this.addNamespaceDecl(XMLSymbols.EMPTY_STRING, XMLSymbols.EMPTY_STRING, element);
                this.fLocalNSBinder.declarePrefix(XMLSymbols.EMPTY_STRING, null);
                this.fNamespaceContext.declarePrefix(XMLSymbols.EMPTY_STRING, null);
            }
        }
        if (attributes != null) {
            attributes.cloneMap(this.fAttributeList);
            for (int i = 0; i < this.fAttributeList.size(); ++i) {
                Attr attr = (Attr)this.fAttributeList.get(i);
                (this.fLocator.fRelatedNode = attr).normalize();
                String value = attr.getValue();
                uri = attr.getNamespaceURI();
                if (value == null) {
                    value = XMLSymbols.EMPTY_STRING;
                }
                if (this.fDocument.errorChecking && (this.fConfiguration.features & 0x100) != 0x0) {
                    isAttrValueWF(this.fErrorHandler, this.fError, this.fLocator, attributes, attr, value, this.fDocument.isXML11Version());
                    if (this.fDocument.isXMLVersionChanged()) {
                        boolean wellformed;
                        if (this.fNamespaceValidation) {
                            wellformed = CoreDocumentImpl.isValidQName(attr.getPrefix(), attr.getLocalName(), this.fDocument.isXML11Version());
                        }
                        else {
                            wellformed = CoreDocumentImpl.isXMLName(attr.getNodeName(), this.fDocument.isXML11Version());
                        }
                        if (!wellformed) {
                            String msg3 = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "wf-invalid-character-in-node-name", new Object[] { "Attr", attr.getNodeName() });
                            reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg3, (short)2, "wf-invalid-character-in-node-name");
                        }
                    }
                }
                if (uri != null) {
                    prefix = attr.getPrefix();
                    prefix = ((prefix == null || prefix.length() == 0) ? XMLSymbols.EMPTY_STRING : this.fSymbolTable.addSymbol(prefix));
                    this.fSymbolTable.addSymbol(attr.getLocalName());
                    if (uri == null || !uri.equals(NamespaceContext.XMLNS_URI)) {
                        ((AttrImpl)attr).setIdAttribute(false);
                        uri = this.fSymbolTable.addSymbol(uri);
                        String declaredURI = this.fNamespaceContext.getURI(prefix);
                        if (prefix == XMLSymbols.EMPTY_STRING || declaredURI != uri) {
                            String declaredPrefix = this.fNamespaceContext.getPrefix(uri);
                            if (declaredPrefix != null && declaredPrefix != XMLSymbols.EMPTY_STRING) {
                                prefix = declaredPrefix;
                            }
                            else {
                                if (prefix == XMLSymbols.EMPTY_STRING || this.fLocalNSBinder.getURI(prefix) != null) {
                                    int counter;
                                    for (counter = 1, prefix = this.fSymbolTable.addSymbol("NS" + counter++); this.fLocalNSBinder.getURI(prefix) != null; prefix = this.fSymbolTable.addSymbol("NS" + counter++)) {}
                                }
                                this.addNamespaceDecl(prefix, uri, element);
                                value = this.fSymbolTable.addSymbol(value);
                                this.fLocalNSBinder.declarePrefix(prefix, value);
                                this.fNamespaceContext.declarePrefix(prefix, uri);
                            }
                            attr.setPrefix(prefix);
                        }
                    }
                }
                else {
                    ((AttrImpl)attr).setIdAttribute(false);
                    if (attr.getLocalName() == null) {
                        if (this.fNamespaceValidation) {
                            String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "NullLocalAttrName", new Object[] { attr.getNodeName() });
                            reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg, (short)3, "NullLocalAttrName");
                        }
                        else {
                            String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "NullLocalAttrName", new Object[] { attr.getNodeName() });
                            reportDOMError(this.fErrorHandler, this.fError, this.fLocator, msg, (short)2, "NullLocalAttrName");
                        }
                    }
                }
            }
        }
    }
    
    protected final void addNamespaceDecl(String prefix, String uri, ElementImpl element) {
        if (prefix == XMLSymbols.EMPTY_STRING) {
            element.setAttributeNS(NamespaceContext.XMLNS_URI, XMLSymbols.PREFIX_XMLNS, uri);
        }
        else {
            element.setAttributeNS(NamespaceContext.XMLNS_URI, "xmlns:" + prefix, uri);
        }
    }
    
    public static final void isCDataWF(DOMErrorHandler errorHandler, DOMErrorImpl error, DOMLocatorImpl locator, String datavalue, boolean isXML11Version) {
        if (datavalue == null || datavalue.length() == 0) {
            return;
        }
        char[] dataarray = datavalue.toCharArray();
        int datalength = dataarray.length;
        if (isXML11Version) {
            int i = 0;
            while (i < datalength) {
                char c = dataarray[i++];
                if (XML11Char.isXML11Invalid((int)c)) {
                    if (XMLChar.isHighSurrogate((int)c) && i < datalength) {
                        char c2 = dataarray[i++];
                        if (XMLChar.isLowSurrogate((int)c2) && XMLChar.isSupplemental(XMLChar.supplemental(c, c2))) {
                            continue;
                        }
                    }
                    String msg2 = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "InvalidCharInCDSect", new Object[] { Integer.toString((int)c, 16) });
                    reportDOMError(errorHandler, error, locator, msg2, (short)2, "wf-invalid-character");
                }
                else if (c == ']') {
                    int count = i;
                    if (count < datalength && dataarray[count] == ']') {
                        while (++count < datalength && dataarray[count] == ']') {}
                        if (count < datalength && dataarray[count] == '>') {
                            String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "CDEndInContent", (Object[])null);
                            reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
        }
        else {
            int i = 0;
            while (i < datalength) {
                char c = dataarray[i++];
                if (XMLChar.isInvalid((int)c)) {
                    if (XMLChar.isHighSurrogate((int)c) && i < datalength) {
                        char c2 = dataarray[i++];
                        if (XMLChar.isLowSurrogate((int)c2) && XMLChar.isSupplemental(XMLChar.supplemental(c, c2))) {
                            continue;
                        }
                    }
                    String msg2 = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "InvalidCharInCDSect", new Object[] { Integer.toString((int)c, 16) });
                    reportDOMError(errorHandler, error, locator, msg2, (short)2, "wf-invalid-character");
                }
                else if (c == ']') {
                    int count = i;
                    if (count < datalength && dataarray[count] == ']') {
                        while (++count < datalength && dataarray[count] == ']') {}
                        if (count < datalength && dataarray[count] == '>') {
                            String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "CDEndInContent", (Object[])null);
                            reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
        }
        return;
    }
    
    public static final void isXMLCharWF(DOMErrorHandler errorHandler, DOMErrorImpl error, DOMLocatorImpl locator, String datavalue, boolean isXML11Version) {
        if (datavalue == null || datavalue.length() == 0) {
            return;
        }
        char[] dataarray = datavalue.toCharArray();
        int datalength = dataarray.length;
        if (isXML11Version) {
            int i = 0;
            while (i < datalength) {
                if (!XML11Char.isXML11Invalid((int)dataarray[i++])) {
                    continue;
                }
                char ch = dataarray[i - 1];
                if (XMLChar.isHighSurrogate((int)ch) && i < datalength) {
                    char ch2 = dataarray[i++];
                    if (XMLChar.isLowSurrogate((int)ch2) && XMLChar.isSupplemental(XMLChar.supplemental(ch, ch2))) {
                        continue;
                    }
                }
                String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "InvalidXMLCharInDOM", new Object[] { Integer.toString((int)dataarray[i - 1], 16) });
                reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
            }
        }
        else {
            int i = 0;
            while (i < datalength) {
                if (!XMLChar.isInvalid((int)dataarray[i++])) {
                    continue;
                }
                char ch = dataarray[i - 1];
                if (XMLChar.isHighSurrogate((int)ch) && i < datalength) {
                    char ch2 = dataarray[i++];
                    if (XMLChar.isLowSurrogate((int)ch2) && XMLChar.isSupplemental(XMLChar.supplemental(ch, ch2))) {
                        continue;
                    }
                }
                String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "InvalidXMLCharInDOM", new Object[] { Integer.toString((int)dataarray[i - 1], 16) });
                reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
            }
        }
        return;
    }
    
    public static final void isCommentWF(DOMErrorHandler errorHandler, DOMErrorImpl error, DOMLocatorImpl locator, String datavalue, boolean isXML11Version) {
        if (datavalue == null || datavalue.length() == 0) {
            return;
        }
        char[] dataarray = datavalue.toCharArray();
        int datalength = dataarray.length;
        if (isXML11Version) {
            int i = 0;
            while (i < datalength) {
                char c = dataarray[i++];
                if (XML11Char.isXML11Invalid((int)c)) {
                    if (XMLChar.isHighSurrogate((int)c) && i < datalength) {
                        char c2 = dataarray[i++];
                        if (XMLChar.isLowSurrogate((int)c2) && XMLChar.isSupplemental(XMLChar.supplemental(c, c2))) {
                            continue;
                        }
                    }
                    String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "InvalidCharInComment", new Object[] { Integer.toString((int)dataarray[i - 1], 16) });
                    reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                }
                else if (c == '-' && i < datalength && dataarray[i] == '-') {
                    String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "DashDashInComment", (Object[])null);
                    reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                }
                else {
                    continue;
                }
            }
        }
        else {
            int i = 0;
            while (i < datalength) {
                char c = dataarray[i++];
                if (XMLChar.isInvalid((int)c)) {
                    if (XMLChar.isHighSurrogate((int)c) && i < datalength) {
                        char c2 = dataarray[i++];
                        if (XMLChar.isLowSurrogate((int)c2) && XMLChar.isSupplemental(XMLChar.supplemental(c, c2))) {
                            continue;
                        }
                    }
                    String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "InvalidCharInComment", new Object[] { Integer.toString((int)dataarray[i - 1], 16) });
                    reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                }
                else if (c == '-' && i < datalength && dataarray[i] == '-') {
                    String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/TR/1998/REC-xml-19980210", "DashDashInComment", (Object[])null);
                    reportDOMError(errorHandler, error, locator, msg, (short)2, "wf-invalid-character");
                }
                else {
                    continue;
                }
            }
        }
        return;
    }
    
    public static final void isAttrValueWF(DOMErrorHandler errorHandler, DOMErrorImpl error, DOMLocatorImpl locator, NamedNodeMap attributes, Attr a, String value, boolean xml11Version) {
        if (a instanceof AttrImpl && ((AttrImpl)a).hasStringValue()) {
            isXMLCharWF(errorHandler, error, locator, value, xml11Version);
        }
        else {
            NodeList children = a.getChildNodes();
            for (int j = 0; j < children.getLength(); ++j) {
                Node child = children.item(j);
                if (child.getNodeType() == 5) {
                    Document owner = a.getOwnerDocument();
                    Entity ent = null;
                    if (owner != null) {
                        DocumentType docType = owner.getDoctype();
                        if (docType != null) {
                            NamedNodeMap entities = docType.getEntities();
                            ent = (Entity)entities.getNamedItemNS("*", child.getNodeName());
                        }
                    }
                    if (ent == null) {
                        String msg = DOMMessageFormatter.formatMessage("http://www.w3.org/dom/DOMTR", "UndeclaredEntRefInAttrValue", new Object[] { a.getNodeName() });
                        reportDOMError(errorHandler, error, locator, msg, (short)2, "UndeclaredEntRefInAttrValue");
                    }
                }
                else {
                    isXMLCharWF(errorHandler, error, locator, child.getNodeValue(), xml11Version);
                }
            }
        }
    }
    
    public static final void reportDOMError(DOMErrorHandler errorHandler, DOMErrorImpl error, DOMLocatorImpl locator, String message, short severity, String type) {
        if (errorHandler != null) {
            error.reset();
            error.fMessage = message;
            error.fSeverity = severity;
            error.fLocator = locator;
            error.fType = type;
            error.fRelatedData = locator.fRelatedNode;
            if (!errorHandler.handleError(error)) {
                throw new AbortException();
            }
        }
        if (severity == 3) {
            throw new AbortException();
        }
        return;
    }
    
    protected final void updateQName(Node node, QName qname) {
        String prefix = node.getPrefix();
        String namespace = node.getNamespaceURI();
        String localName = node.getLocalName();
        qname.prefix = ((prefix != null && prefix.length() != 0) ? this.fSymbolTable.addSymbol(prefix) : null);
        qname.localpart = ((localName != null) ? this.fSymbolTable.addSymbol(localName) : null);
        qname.rawname = this.fSymbolTable.addSymbol(node.getNodeName());
        qname.uri = ((namespace != null) ? this.fSymbolTable.addSymbol(namespace) : null);
    }
    
    final String normalizeAttributeValue(String value, Attr attr) {
        if (!attr.getSpecified()) {
            return value;
        }
        int end = value.length();
        if (this.fNormalizedValue.ch.length < end) {
            this.fNormalizedValue.ch = new char[end];
        }
        this.fNormalizedValue.length = 0;
        boolean normalized = false;
        for (int i = 0; i < end; ++i) {
            char c = value.charAt(i);
            if (c == '\t' || c == '\n') {
                this.fNormalizedValue.ch[this.fNormalizedValue.length++] = ' ';
                normalized = true;
            }
            else if (c == '\r') {
                normalized = true;
                this.fNormalizedValue.ch[this.fNormalizedValue.length++] = ' ';
                int next = i + 1;
                if (next < end && value.charAt(next) == '\n') {
                    i = next;
                }
            }
            else {
                this.fNormalizedValue.ch[this.fNormalizedValue.length++] = c;
            }
        }
        if (normalized) {
            value = this.fNormalizedValue.toString();
            attr.setValue(value);
            return value;
        }
        return value;
    }
    
    @Override
    public void startDocument(XMLLocator locator, String encoding, NamespaceContext namespaceContext, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void xmlDecl(String version, String encoding, String standalone, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void doctypeDecl(String rootElement, String publicId, String systemId, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void comment(XMLString text, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void processingInstruction(String target, XMLString data, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void startElement(QName element, XMLAttributes attributes, Augmentations augs) throws XNIException {
        Element currentElement = (Element)this.fCurrentNode;
        for (int attrCount = attributes.getLength(), i = 0; i < attrCount; ++i) {
            attributes.getName(i, this.fAttrQName);
            Attr attr = null;
            attr = currentElement.getAttributeNodeNS(this.fAttrQName.uri, this.fAttrQName.localpart);
            if (attr == null) {
                attr = currentElement.getAttributeNode(this.fAttrQName.rawname);
            }
            AttributePSVI attrPSVI = (AttributePSVI)attributes.getAugmentations(i).getItem("ATTRIBUTE_PSVI");
            if (attrPSVI != null) {
                XSTypeDefinition decl = attrPSVI.getMemberTypeDefinition();
                boolean id = false;
                if (decl != null) {
                    id = ((XSSimpleType)decl).isIDType();
                }
                else {
                    decl = attrPSVI.getTypeDefinition();
                    if (decl != null) {
                        id = ((XSSimpleType)decl).isIDType();
                    }
                }
                if (id) {
                    ((ElementImpl)currentElement).setIdAttributeNode(attr, true);
                }
                if (this.fPSVI) {
                    ((PSVIAttrNSImpl)attr).setPSVI(attrPSVI);
                }
                ((AttrImpl)attr).setType(decl);
                if ((this.fConfiguration.features & 0x2) != 0x0) {
                    String normalizedValue = attrPSVI.getSchemaValue().getNormalizedValue();
                    if (normalizedValue != null) {
                        boolean specified = attr.getSpecified();
                        attr.setValue(normalizedValue);
                        if (!specified) {
                            ((AttrImpl)attr).setSpecified(specified);
                        }
                    }
                }
            }
            else {
                String type = null;
                boolean isDeclared = Boolean.TRUE.equals(attributes.getAugmentations(i).getItem("ATTRIBUTE_DECLARED"));
                if (isDeclared) {
                    type = attributes.getType(i);
                    if ("ID".equals(type)) {
                        ((ElementImpl)currentElement).setIdAttributeNode(attr, true);
                    }
                }
                ((AttrImpl)attr).setType(type);
            }
        }
    }
    
    @Override
    public void emptyElement(QName element, XMLAttributes attributes, Augmentations augs) throws XNIException {
        this.startElement(element, attributes, augs);
        this.endElement(element, augs);
    }
    
    @Override
    public void startGeneralEntity(String name, XMLResourceIdentifier identifier, String encoding, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void textDecl(String version, String encoding, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void endGeneralEntity(String name, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void characters(XMLString text, Augmentations augs) throws XNIException {
    }
    
    @Override
    public void ignorableWhitespace(XMLString text, Augmentations augs) throws XNIException {
        this.fAllWhitespace = true;
    }
    
    @Override
    public void endElement(QName element, Augmentations augs) throws XNIException {
        if (augs != null) {
            ElementPSVI elementPSVI = (ElementPSVI)augs.getItem("ELEMENT_PSVI");
            if (elementPSVI != null) {
                ElementImpl elementNode = (ElementImpl)this.fCurrentNode;
                if (this.fPSVI) {
                    ((PSVIElementNSImpl)this.fCurrentNode).setPSVI(elementPSVI);
                }
                if ((elementNode instanceof ElementNSImpl)) {
                    XSTypeDefinition type = elementPSVI.getMemberTypeDefinition();
                    if (type == null) {
                        type = elementPSVI.getTypeDefinition();
                    }
                    ((ElementNSImpl)elementNode).setType(type);
                }
                String normalizedValue = elementPSVI.getSchemaValue().getNormalizedValue();
                if ((this.fConfiguration.features & 0x2) != 0x0) {
                    if (normalizedValue != null) {
                        elementNode.setTextContent(normalizedValue);
                    }
                }
                else {
                    String text = elementNode.getTextContent();
                    if (text.length() == 0 && normalizedValue != null) {
                        elementNode.setTextContent(normalizedValue);
                    }
                }
                return;
            }
        }
        if ((this.fCurrentNode instanceof ElementNSImpl)) {
            ((ElementNSImpl)this.fCurrentNode).setType(null);
        }
    }
    
    @Override
    public void startCDATA(Augmentations augs) throws XNIException {
    }
    
    @Override
    public void endCDATA(Augmentations augs) throws XNIException {
    }
    
    @Override
    public void endDocument(Augmentations augs) throws XNIException {
    }
    
    @Override
    public void setDocumentSource(XMLDocumentSource source) {
    }
    
    @Override
    public XMLDocumentSource getDocumentSource() {
        return null;
    }
    
    static {
        EMPTY_STRING = new XMLString();
    }
    
    protected final class XMLAttributesProxy implements XMLAttributes
    {
        protected AttributeMap fAttributes;
        protected CoreDocumentImpl fDocument;
        protected ElementImpl fElement;
        protected Vector<String> fDTDTypes;
        protected final Vector<Augmentations> fAugmentations;
        
        protected XMLAttributesProxy() {
            this.fDTDTypes = new Vector(5);
            this.fAugmentations = new Vector(5);
        }
        
        public void setAttributes(AttributeMap attributes, CoreDocumentImpl doc, ElementImpl elem) {
            this.fDocument = doc;
            this.fAttributes = attributes;
            this.fElement = elem;
            if (attributes != null) {
                int length = attributes.getLength();
                this.fDTDTypes.setSize(length);
                this.fAugmentations.setSize(length);
                for (int i = 0; i < length; ++i) {
                    this.fAugmentations.setElementAt(new AugmentationsImpl(), i);
                }
            }
            else {
                this.fDTDTypes.setSize(0);
                this.fAugmentations.setSize(0);
            }
        }
        
        @Override
        public int addAttribute(QName qname, String attrType, String attrValue) {
            int index = this.fElement.getXercesAttribute(qname.uri, qname.localpart);
            if (index < 0) {
                AttrImpl attr = (AttrImpl)((CoreDocumentImpl)this.fElement.getOwnerDocument()).createAttributeNS(qname.uri, qname.rawname, qname.localpart);
                attr.setNodeValue(attrValue);
                index = this.fElement.setXercesAttributeNode(attr);
                this.fDTDTypes.insertElementAt(attrType, index);
                this.fAugmentations.insertElementAt(new AugmentationsImpl(), index);
                attr.setSpecified(false);
            }
            return index;
        }
        
        @Override
        public void removeAllAttributes() {
        }
        
        @Override
        public void removeAttributeAt(int attrIndex) {
        }
        
        @Override
        public int getLength() {
            return (this.fAttributes != null) ? this.fAttributes.getLength() : 0;
        }
        
        @Override
        public int getIndex(String qName) {
            return -1;
        }
        
        @Override
        public int getIndex(String uri, String localPart) {
            return -1;
        }
        
        @Override
        public void setName(int attrIndex, QName attrName) {
        }
        
        @Override
        public void getName(int attrIndex, QName attrName) {
            if (this.fAttributes != null) {
                DOMNormalizer.this.updateQName((Node)this.fAttributes.getItem(attrIndex), attrName);
            }
        }
        
        @Override
        public String getPrefix(int index) {
            if (this.fAttributes != null) {
                Node node = (Node)this.fAttributes.getItem(index);
                String prefix = node.getPrefix();
                prefix = ((prefix != null && prefix.length() != 0) ? DOMNormalizer.this.fSymbolTable.addSymbol(prefix) : null);
                return prefix;
            }
            return null;
        }
        
        @Override
        public String getURI(int index) {
            if (this.fAttributes != null) {
                Node node = (Node)this.fAttributes.getItem(index);
                String namespace = node.getNamespaceURI();
                namespace = ((namespace != null) ? DOMNormalizer.this.fSymbolTable.addSymbol(namespace) : null);
                return namespace;
            }
            return null;
        }
        
        @Override
        public String getLocalName(int index) {
            if (this.fAttributes != null) {
                Node node = (Node)this.fAttributes.getItem(index);
                String localName = node.getLocalName();
                localName = ((localName != null) ? DOMNormalizer.this.fSymbolTable.addSymbol(localName) : null);
                return localName;
            }
            return null;
        }
        
        @Override
        public String getQName(int index) {
            if (this.fAttributes != null) {
                Node node = (Node)this.fAttributes.getItem(index);
                String rawname = DOMNormalizer.this.fSymbolTable.addSymbol(node.getNodeName());
                return rawname;
            }
            return null;
        }
        
        @Override
        public QName getQualifiedName(int index) {
            return null;
        }
        
        @Override
        public void setType(int attrIndex, String attrType) {
            this.fDTDTypes.setElementAt(attrType, attrIndex);
        }
        
        @Override
        public String getType(int index) {
            String type = (String)this.fDTDTypes.elementAt(index);
            return (type != null) ? this.getReportableType(type) : "CDATA";
        }
        
        @Override
        public String getType(String qName) {
            return "CDATA";
        }
        
        @Override
        public String getType(String uri, String localName) {
            return "CDATA";
        }
        
        private String getReportableType(String type) {
            if (type.charAt(0) == '(') {
                return "NMTOKEN";
            }
            return type;
        }
        
        @Override
        public void setValue(int attrIndex, String attrValue) {
            if (this.fAttributes != null) {
                AttrImpl attr = (AttrImpl)this.fAttributes.getItem(attrIndex);
                boolean specified = attr.getSpecified();
                attr.setValue(attrValue);
                attr.setSpecified(specified);
            }
        }
        
        @Override
        public void setValue(int attrIndex, String attrValue, XMLString value) {
            this.setValue(attrIndex, value.toString());
        }
        
        @Override
        public String getValue(int index) {
            return (this.fAttributes != null) ? this.fAttributes.item(index).getNodeValue() : "";
        }
        
        @Override
        public String getValue(String qName) {
            return null;
        }
        
        @Override
        public String getValue(String uri, String localName) {
            if (this.fAttributes != null) {
                Node node = this.fAttributes.getNamedItemNS(uri, localName);
                return (node != null) ? node.getNodeValue() : null;
            }
            return null;
        }
        
        @Override
        public void setNonNormalizedValue(int attrIndex, String attrValue) {
        }
        
        @Override
        public String getNonNormalizedValue(int attrIndex) {
            return null;
        }
        
        @Override
        public void setSpecified(int attrIndex, boolean specified) {
            AttrImpl attr = (AttrImpl)this.fAttributes.getItem(attrIndex);
            attr.setSpecified(specified);
        }
        
        @Override
        public boolean isSpecified(int attrIndex) {
            return ((Attr)this.fAttributes.getItem(attrIndex)).getSpecified();
        }
        
        @Override
        public Augmentations getAugmentations(int attributeIndex) {
            return (Augmentations)this.fAugmentations.elementAt(attributeIndex);
        }
        
        @Override
        public Augmentations getAugmentations(String uri, String localPart) {
            return null;
        }
        
        @Override
        public Augmentations getAugmentations(String qName) {
            return null;
        }
        
        @Override
        public void setAugmentations(int attrIndex, Augmentations augs) {
            this.fAugmentations.setElementAt(augs, attrIndex);
        }
    }
}
